/**
  @file d2xx_loopback_test1.c
  @brief
  USB D2XX Internal Loopback Test.

  Any character typed on D2XX Ports will be echoed back on the D2XX Ports.
 */
/*
 * ============================================================================
 * History
 * =======
 * UKB 20 Jan 2016 : Created
 *
 * (C) Copyright 2016, Future Technology Devices International Ltd.
 * ============================================================================
 *
 * This source code ("the Software") is provided by Future Technology Devices
 * International Limited ("FTDI") subject to the licence terms set out
 * http://www.ftdichip.com/FTSourceCodeLicenceTerms.htm ("the Licence Terms").
 * You must read the Licence Terms before downloading or using the Software.
 * By installing or using the Software you agree to the Licence Terms. If you
 * do not agree to the Licence Terms then do not download or use the Software.
 *
 * Without prejudice to the Licence Terms, here is a summary of some of the key
 * terms of the Licence Terms (and in the event of any conflict between this
 * summary and the Licence Terms then the text of the Licence Terms will
 * prevail).
 *
 * The Software is provided "as is".
 * There are no warranties (or similar) in relation to the quality of the
 * Software. You use it at your own risk.
 * The Software should not be used in, or for, any medical device, system or
 * appliance. There are exclusions of FTDI liability for certain types of loss
 * such as: special loss or damage; incidental loss or damage; indirect or
 * consequential loss or damage; loss of income; loss of business; loss of
 * profits; loss of revenue; loss of contracts; business interruption; loss of
 * the use of money or anticipated savings; loss of information; loss of
 * opportunity; loss of goodwill or reputation; and/or loss of, damage to or
 * corruption of data.
 * There is a monetary cap on FTDI's liability.
 * The Software may have subsequently been amended by another user and then
 * distributed by that other user ("Adapted Software").  If so that user may
 * have additional licence terms that apply to those amendments. However, FTDI
 * has no liability in relation to those amendments.
 * ============================================================================
 */

/* INCLUDES ************************************************************************/

#include <stdint.h>
#include <string.h>
#include <ctype.h>

#include "ftd2xx_api.h"
#include <ft900_startup_dfu.h>
#include <ft900.h>


/* UART support for printf output. */
#include "tinyprintf.h"

/* CONSTANTS ***********************************************************************/

#define GPIO_REMOTE_WAKEUP

#ifdef GPIO_REMOTE_WAKEUP
	//GPIO30
#define GPIO_PIN  18
#endif

#define MAX_READBUFFER_SIZE  1024
#define MAX_D2XX_INTERFACES  2

/* GLOBAL VARIABLES ****************************************************************/

extern TD2XX_DeviceConfiguration __pD2XXDefaultConfiguration[]; // pointer to D2XX Config in flash

/* LOCAL VARIABLES *****************************************************************/

static uint8_t D2XXTest_ReadBuffer[MAX_READBUFFER_SIZE];
static char *D2XXTest_EventStrings[D2XX_EVT_DFU_DETACH+1] = {	"D2XX_EVT_SUSPEND",    	/**< SUSPEND EVENT from USB Host */
																	"D2XX_EVT_RESUME",	 	/**< RESUME EVENT from USB Host */
																	"D2XX_EVT_RESET",	 		/**< RESET Vendor Command from D2XX Application */
																	"D2XX_EVT_READY",	 		/**< D2XX enters Ready state where READ/WRITE requests are processed*/
																	"D2XX_EVT_UNREADY",	 	/**< D2XX exits Ready state*/
																	"D2XX_EVT_DFU_DETACH"		/**< DFU DETACH Class command from DFU application */
																};
static volatile int D2XXTEST__DfuDetach = 0;
static volatile int D2XXTEST__Ready = 0;
static volatile uint8_t SetRemoteWakeup = 0;
static uint8_t RemoteWakeupEnable = 0;
#if 1
static uint8_t D2XXTEST__Suspend = 0;
static volatile uint8_t D2XXTEST__Wakeup = 0;
static volatile uint8_t wkupSource = 0;
#endif

/* LOCAL FUNCTIONS / INLINES *******************************************************/

void setup(void);
void loop (void);
void debug_uart_init(void);
void tfp_putc(void* p, char c);
void d2xx_callback(ED2XX_EventCode  eventID, void *ref, void* param1, void* param2);
void ISR_powermanagement(void);
#ifdef GPIO_REMOTE_WAKEUP
void ISR_gpio();
#endif

#undef BDFU  /* Undefine for C-DFU */

#ifdef BDFU
extern char __PMSIZE;
typedef void (*__dfumain_t)(unsigned int, unsigned int) __attribute__((noreturn));
#define GO_BDFU(VID, PID, REL) ((__dfumain_t)(&__PMSIZE - 12))( (PID) << 16 | (VID), REL);
#endif

/* FUNCTIONS ***********************************************************************/
int main(void)
{

	/* Setup UART and D2XX */
    setup();

    for(;;){

    	/* Test process loop */
		loop();


		/* Event handling */
		if (D2XXTEST__DfuDetach)
		{
			delayms(100);
			/* DFU detach request had come. Exit D2XX */
			D2XX_Exit();

			/* Take the system to DFU Mode. Provide timeout in ms to
			 * wait for re-enumeration with the host
			 * */
#ifndef BDFU
			STARTUP_DFU(0);
#else
			tfp_printf("Switching to DFU");
			if (USBD_DFU_is_runtime()) /* to ward off linkage issue */
			{
				GO_BDFU(0x0403, 0x0FDE, 0x2300); // use default VID, PID and REL
			}
#endif
			break;
		}

		/* Suspend mode handling */
		if (D2XXTEST__Suspend)
		{

			wkupSource = 0;
			tfp_printf("d2xx_loopback_test1: Entering Suspend Mode at Master \r\n");
			delayms(100);
#ifdef GPIO_REMOTE_WAKEUP
			SetRemoteWakeup = 0;
			if (RemoteWakeupEnable)
			{
				tfp_printf("d2xx_loopback_test1: Configuring GPIO pin to wakeup \r\n");
					#if 1
						/* Set up the pin */
						gpio_dir(GPIO_PIN, pad_dir_input);
						gpio_pull(GPIO_PIN, pad_pull_pullup);

						/* Attach an interrupt */
						gpio_interrupt_enable(GPIO_PIN, gpio_int_edge_falling);
						interrupt_attach(interrupt_gpio, (uint8_t)interrupt_gpio, ISR_gpio);
						tfp_printf("d2xx_loopback_test1: GPIO pin setup for interrupt\r\n");
					#endif
			}
#endif

			/* Power down Mode */
			SYS->PMCFG_L |= MASK_SYS_PMCFG_PM_PWRDN_MODE;
			SYS->PMCFG_L |= MASK_SYS_PMCFG_PM_PWRDN;
			asm volatile ("nop");
			asm volatile ("nop");
			asm volatile ("nop");
			asm volatile ("nop");
			asm volatile ("nop");

			//wait for wakeup
			while(!D2XXTEST__Wakeup);

			interrupt_disable_globally();

#ifdef GPIO_REMOTE_WAKEUP
			gpio_interrupt_disable(GPIO_PIN);
			interrupt_detach(interrupt_gpio);
			gpio_is_interrupted(GPIO_PIN);
#endif

			SYS->PMCFG_L &= ~MASK_SYS_PMCFG_PM_PWRDN_MODE;
			SYS->PMCFG_L &= ~MASK_SYS_PMCFG_PM_PWRDN;


			D2XXTEST__Suspend = 0;
			D2XXTEST__Wakeup = 0;
			interrupt_enable_globally();
			tfp_printf("d2xx_loopback_test1: Exiting Suspend Mode at Master: %02X \r\n", wkupSource);
			if (SetRemoteWakeup)
			{
				D2XX_IOCTL(0, D2XX_IOCTL_SYS_REMOTE_WAKEUP, &SetRemoteWakeup, NULL);
				SetRemoteWakeup = 0;
				tfp_printf("d2xx_loopback_test1: Sending remote wakeup from Application \r\n");
			}
		}
    }
    return 0;
}

/* Test function that does the D2XX loopback functionality */
void loop (void)
{
	ED2XX_ErrorCode retVal = 0;
	int bytesWritten;

		for(int i = 1; i <= MAX_D2XX_INTERFACES; i++)
		{
			/*read the bytes out... */
			retVal = D2XX_Read(i, D2XXTest_ReadBuffer,MAX_READBUFFER_SIZE);
			/* Write the byte out... */
			if (retVal > 0)
			{
					int	ptr = 0;
					do
					{
						bytesWritten = D2XX_Write(i, &D2XXTest_ReadBuffer[ptr], retVal);
						if (bytesWritten > 0)
						{
							retVal -= bytesWritten;
							ptr += bytesWritten;
						}
					}while (retVal);
			}
		}
}

void setup(void)
{
	ED2XX_ErrorCode retVal;

	D2XXTEST__Suspend = 0;
	D2XXTEST__Wakeup = 0;

#ifdef BDFU
	timer_prescaler(1000); /* to ward off linkage issue */
#endif
	interrupt_attach(interrupt_0, (int8_t)interrupt_0, ISR_powermanagement);

	// UART initialisation
	debug_uart_init();

    /* Print out a welcome message... */
	tfp_printf("(C) Copyright 2016, Future Technology Devices International Ltd. \r\n"
			"--------------------------------------------------------------------- \r\n"
			"Welcome to D2XX Example 1... \r\n"
			"\r\n"
			"Enter any text on the D2XX port, the same is echoed back on the same port... \r\n"
			"--------------------------------------------------------------------- \r\n"
	);

	sys_enable(sys_device_timer_wdt);

	interrupt_enable_globally(); //needed for interrupts

   	retVal = D2XX_Init(__pD2XXDefaultConfiguration, d2xx_callback, NULL);
	tfp_printf("d2xx_loopback_test1: D2XX_Init() called, Result: %d \r\n", retVal);
}

/** @name tfp_putc
 *  @details Machine dependent putc function for tfp_printf (tinyprintf) library.
 *  @param p Parameters (machine dependent)
 *  @param c The character to write
 */
void tfp_putc(void* p, char c)
{
	uart_write((ft900_uart_regs_t*)p, (uint8_t)c);
}

/** @name d2xx_callback
 *  @details Callback function to the D2XX library.
 *  @param
 *  @param
 */
void d2xx_callback(ED2XX_EventCode  eventID, void *ref, void* param1, void* param2)
{
	tfp_printf("d2xx_loopback_test1: %s is received \r\n", D2XXTest_EventStrings[eventID]);
	if (D2XXTEST__Suspend)
	{
		wkupSource |= 0x10;
		D2XXTEST__Wakeup = 1;
	}
	if ((eventID == D2XX_EVT_SUSPEND) || (eventID == D2XX_EVT_UNREADY))
	{
		D2XXTEST__Suspend = 1;
		D2XXTEST__Wakeup = 0;
		if ((eventID == D2XX_EVT_SUSPEND) && (param1))
		{
			RemoteWakeupEnable = *((uint8_t *)param1);
			tfp_printf("d2xx_loopback_test1: RemoteWakeup Enabled :%d \r\n", RemoteWakeupEnable);
		}
	}
	else if (eventID == D2XX_EVT_DFU_DETACH)
	{
		D2XXTEST__DfuDetach = 1;
	}
	else if (eventID == D2XX_EVT_READY)
	{
		D2XXTEST__Ready = 1;
	}

}

/* Initialize the UART for the testing */
void debug_uart_init(void)
{

	/* Enable the UART Device... */
	sys_enable(sys_device_uart0);
	/* Make GPIO48 function as UART0_TXD and GPIO49 function as UART0_RXD... */
	gpio_function(48, pad_uart0_txd); /* UART0 TXD */
	gpio_function(49, pad_uart0_rxd); /* UART0 RXD */

	// Open the UART using the coding required.
	uart_open(UART0,                    /* Device */
			1,                        /* Prescaler = 1 */
			UART_DIVIDER_19200_BAUD,  /* Divider = 1302 */
			uart_data_bits_8,         /* No. buffer Bits */
			uart_parity_none,         /* Parity */
			uart_stop_bits_1);        /* No. Stop Bits */

	/* Print out a welcome message... */
	uart_puts(UART0,
			"\x1B[2J" /* ANSI/VT100 - Clear the Screen */
			"\x1B[H"  /* ANSI/VT100 - Move Cursor to Home */
	);

	/* Enable tfp_printf() functionality... */
	init_printf(UART0, tfp_putc);
}


/* Power management ISR */
void ISR_powermanagement(void)
{


	if (SYS->PMCFG_H & MASK_SYS_PMCFG_PM_GPIO_IRQ_PEND)
	{
		// Clear GPIO wakeup pending
		SYS->PMCFG_H = MASK_SYS_PMCFG_PM_GPIO_IRQ_PEND;
		if (D2XXTEST__Suspend)
		{
			wkupSource |= 2;
			D2XXTEST__Wakeup = 1;
		}
	}

	if (SYS->PMCFG_H & MASK_SYS_PMCFG_DEV_CONN_DEV)
	{
		// Clear connection interrupt
		SYS->PMCFG_H = MASK_SYS_PMCFG_DEV_CONN_DEV;

	}

	if (SYS->PMCFG_H & MASK_SYS_PMCFG_DEV_DIS_DEV)
	{
		// Clear disconnection interrupt
		SYS->PMCFG_H = MASK_SYS_PMCFG_DEV_DIS_DEV;

	}

	if (SYS->PMCFG_H & MASK_SYS_PMCFG_HOST_RST_DEV)
	{
		// Clear Host Reset interrupt
		SYS->PMCFG_H = MASK_SYS_PMCFG_HOST_RST_DEV;
		USBD_wakeup();
	}

	if (SYS->PMCFG_H & MASK_SYS_PMCFG_HOST_RESUME_DEV)
	{
		// Clear Host Resume interrupt
		SYS->PMCFG_H = MASK_SYS_PMCFG_HOST_RESUME_DEV;
		if(! (SYS->MSC0CFG & MASK_SYS_MSC0CFG_DEV_RMWAKEUP))
		{
			// If we are driving K-state on Device USB port;
			// We must maintain the 1ms requirement before resuming the phy
			USBD_wakeup();
		}
	}
}


#ifdef GPIO_REMOTE_WAKEUP
void ISR_gpio()
{
    if (gpio_is_interrupted(GPIO_PIN))
    {
    	tfp_printf("d2xx_loopback_test1: gpio interrupt \r\n");
		if (D2XXTEST__Suspend)
		{
			D2XXTEST__Wakeup = 1;
			SetRemoteWakeup = 1;
		}
    }
}
#endif
