/**
  @file usbh_cdcacm_test_1.c
  @brief
  USB CDC ACM Device Test

  This example program will bridge data from the UART to a CDC ACM device
  on the USB.
 */
/*
 * ============================================================================
 * History
 * =======
 * GDM 21 Jan 2015 : Created
 *
 * (C) Copyright 2016, Future Technology Devices International Ltd.
 * ============================================================================
 *
 * This source code ("the Software") is provided by Future Technology Devices
 * International Limited ("FTDI") subject to the licence terms set out
 * http://www.ftdichip.com/FTSourceCodeLicenceTerms.htm ("the Licence Terms").
 * You must read the Licence Terms before downloading or using the Software.
 * By installing or using the Software you agree to the Licence Terms. If you
 * do not agree to the Licence Terms then do not download or use the Software.
 *
 * Without prejudice to the Licence Terms, here is a summary of some of the key
 * terms of the Licence Terms (and in the event of any conflict between this
 * summary and the Licence Terms then the text of the Licence Terms will
 * prevail).
 *
 * The Software is provided "as is".
 * There are no warranties (or similar) in relation to the quality of the
 * Software. You use it at your own risk.
 * The Software should not be used in, or for, any medical device, system or
 * appliance. There are exclusions of FTDI liability for certain types of loss
 * such as: special loss or damage; incidental loss or damage; indirect or
 * consequential loss or damage; loss of income; loss of business; loss of
 * profits; loss of revenue; loss of contracts; business interruption; loss of
 * the use of money or anticipated savings; loss of information; loss of
 * opportunity; loss of goodwill or reputation; and/or loss of, damage to or
 * corruption of data.
 * There is a monetary cap on FTDI's liability.
 * The Software may have subsequently been amended by another user and then
 * distributed by that other user ("Adapted Software").  If so that user may
 * have additional licence terms that apply to those amendments. However, FTDI
 * has no liability in relation to those amendments.
 * ============================================================================
 */

/* INCLUDES ************************************************************************/

#include <stdint.h>
#include <string.h>
#include <ctype.h>

#include <ft900.h>
#include <ft900_usb.h>
#include <ft900_usb_cdc.h>
#include <ft900_usb_dfu.h>
#include <ft900_usbd_dfu.h>

/* UART support for printf output. */
#include "tinyprintf.h"

/* For MikroC const qualifier will place variables in Flash
 * not just make them constant.
 */
#if defined(__GNUC__)
#define DESCRIPTOR_QUALIFIER const
#elif defined(__MIKROC_PRO_FOR_FT90x__)
#define DESCRIPTOR_QUALIFIER data
#endif

/* CONSTANTS ***********************************************************************/

/**
 @name USB and Hub Configuration
 @brief Indication of how the USB device is powered and the size of
 	 the control endpoint Max Packets.
 */
//@{
// USB Bus Powered - set to 1 for self-powered or 0 for bus-powered
#ifndef USB_SELF_POWERED
#define USB_SELF_POWERED 0
#endif // USB_SELF_POWERED
#if USB_SELF_POWERED == 1
#define USB_CONFIG_BMATTRIBUTES_VALUE (USB_CONFIG_BMATTRIBUTES_SELF_POWERED | USB_CONFIG_BMATTRIBUTES_RESERVED_SET_TO_1)
#else // USB_SELF_POWERED
#define USB_CONFIG_BMATTRIBUTES_VALUE USB_CONFIG_BMATTRIBUTES_RESERVED_SET_TO_1
#endif // USB_SELF_POWERED
// USB Endpoint Zero packet size (both must match)
#define USB_CONTROL_EP_MAX_PACKET_SIZE 64
#define USB_CONTROL_EP_SIZE USBD_EP_SIZE_64
//@}


/**
 @name DFU Configuration
 @brief Determines the parts of the DFU specification which are supported
        by the DFU library code. Features can be disabled if required.
 */
//@{
#define DFU_ATTRIBUTES USBD_DFU_ATTRIBUTES
//@}

/**
 @name Device Configuration Areas
 @brief Size and location reserved for string descriptors.
 Leaving the allocation size blank will make an array exactly the size
 of the string allocation.
 Note: Specifying the location is not supported by the GCC compiler.
 */
//@{
// String descriptors - allow a maximum of 256 bytes for this
#define STRING_DESCRIPTOR_LOCATION 0x80
#define STRING_DESCRIPTOR_ALLOCATION 0x100
//@}

/**
 @name DFU_TRANSFER_SIZE definition.
 @brief Number of bytes in block, sent in each DFU_DNLOAD request
 from the DFU update program on the host. This is simplified
 in that the meaning of a block is an arbitrary number of
 bytes. This is intentionally a multiple of the maximum
 packet size for the control endpoints.
 It is used in the DFU functional descriptor as wTransferSize.
 The maximum size supported by the DFU library is 256 bytes
 which is the size of a page of Flash.
 */
//@{
#define DFU_TRANSFER_SIZE USBD_DFU_MAX_BLOCK_SIZE
#define DFU_TIMEOUT USBD_DFU_TIMEOUT
//@}

/**
 @name USB_PID_CDCACM configuration.
 @brief Run Time Product ID for CDC function.
 */
//@{
#define USB_PID_CDCACM 0x0fd1
//@}

/**
 @name DFU_USB_PID_DFUMODE configuration.
 @brief FTDI predefined DFU Mode Product ID.
 */
//@{
#define DFU_USB_PID_DFUMODE 0x0fde
//@}

/**
 @name DFU_USB_INTERFACE configuration..
 @brief Run Time and DFU Mode Interface Numbers.
 */
//@{
#define DFU_USB_INTERFACE_RUNTIME 2
#define DFU_USB_INTERFACE_DFUMODE 0
//@}

/**
 @name WCID_VENDOR_REQUEST_CODE for WCID.
 @brief Unique vendor request code for WCID OS Vendor Extension validation.
*/
//@{
#define WCID_VENDOR_REQUEST_CODE	 0xF1
//@}

/**
 @brief Endpoint definitions for CDCACM.
 */
//@{
#define CDC_EP_NOTIFICATION 			USBD_EP_1
#define CDC_EP_DATA_IN 					USBD_EP_2
#define CDC_EP_DATA_OUT 				USBD_EP_3
#define CDC_NOTIFICATION_EP_SIZE 		0x10
#define CDC_NOTIFICATION_USBD_EP_SIZE 	USBD_EP_SIZE_16
#define CDC_DATA_EP_SIZE				0x40
#define CDC_DATA_USBD_EP_SIZE 			USBD_EP_SIZE_64
//@}

/**
 @brief Change this value in order to modify the size of the Tx and Rx ring buffers
  used to implement UART buffering. Make the ring buffer size 2^n for optimumisation.
 */
#define RINGBUFFER_SIZE (128)

/* GLOBAL VARIABLES ****************************************************************/

/* LOCAL VARIABLES *****************************************************************/

/**
 @name string_descriptor
 @brief Table of USB String descriptors

 This is placed at a fixed location in the const section allowing
 up-to 256 bytes of string descriptors to be defined. These can be
 modified or replaced by a utility or binary editor without
 requiring a recompilation of the firmware.
 They are placed at offset 0x100 and reserve 0x100 bytes.
 The data is not stored in section and is therefore
 regarded as const.
 */
DESCRIPTOR_QUALIFIER /*__at(STRING_DESCRIPTOR_LOCATION)*/ uint8_t string_descriptor[STRING_DESCRIPTOR_ALLOCATION] =
{
		// String 0 is actually a list of language IDs supported by the real strings.
		0x04, USB_DESCRIPTOR_TYPE_STRING, 0x09,
		0x04, // 0409 = English (US)

		// String 1 (Manufacturer): "FTDI"
		0x0a, USB_DESCRIPTOR_TYPE_STRING, 'F', 0x00, 'T', 0x00, 'D', 0x00, 'I',
		0x00,

		// String 2 (Product): "FT900 CDCACM"
		0x1C, USB_DESCRIPTOR_TYPE_STRING, 'F', 0x00, 'T', 0x00, '9', 0x00, '0',
		0x00, '0', 0x00, ' ', 0x00, 'C', 0x00, 'D', 0x00, 'C', 0x00, ' ', 0x00,
		'A', 0x00, 'C', 0x00, 'M', 0x00,

		// String 3 (Serial Number): "FT424242"
		0x12, USB_DESCRIPTOR_TYPE_STRING, 'F', 0x00, 'T', 0x00, '4', 0x00, '2',
		0x00, '4', 0x00, '2', 0x00, '4', 0x00, '2', 0x00,

		// String 4 (DFU Product Name): "FT900 DFU Mode"
		0x1E, USB_DESCRIPTOR_TYPE_STRING, 'F', 0x00, 'T', 0x00, '9', 0x00, '0',
		0x00, '0', 0x00, ' ', 0x00, 'D', 0x00, 'F', 0x00, 'U', 0x00, ' ', 0x00,
		'M', 0x00, 'o', 0x00, 'd', 0x00, 'e', 0x00,

		// String 5 (Interface Name): "DFU Interface"
		0x1c, USB_DESCRIPTOR_TYPE_STRING, 'D', 0x00, 'F', 0x00, 'U', 0x00, ' ',
		0x00, 'I', 0x00, 'n', 0x00, 't', 0x00, 'e', 0x00, 'r', 0x00, 'f', 0x00,
		'a', 0x00, 'c', 0x00, 'e', 0x00,

		// END OF STRINGS
		0x00
};

/**
 @name wcid_string_descriptor
 @brief USB String descriptor for WCID identification.
 */
DESCRIPTOR_QUALIFIER uint8_t wcid_string_descriptor[USB_MICROSOFT_WCID_STRING_LENGTH] = {
		USB_MICROSOFT_WCID_STRING(WCID_VENDOR_REQUEST_CODE)
};

/**
 @name device_descriptor_cdcacm
 @brief Device descriptor for Run Time mode.
 */
DESCRIPTOR_QUALIFIER USB_device_descriptor device_descriptor_cdcacm =
{
		0x12, /* bLength */
		USB_DESCRIPTOR_TYPE_DEVICE, /* bDescriptorType */
		USB_BCD_VERSION_2_0, /* bcdUSB */          // V2.0
		USB_CLASS_DEVICE, /* bDeviceClass */       // Defined in interface
		USB_SUBCLASS_DEVICE, /* bDeviceSubClass */ // Defined in interface
		USB_PROTOCOL_DEVICE, /* bDeviceProtocol */ // Defined in interface
		USB_CONTROL_EP_MAX_PACKET_SIZE, /* bMaxPacketSize0 */
		USB_VID_FTDI, /* idVendor */   // idVendor: 0x0403 (FTDI)
		USB_PID_CDCACM, /* idProduct */ // idProduct: 0x0fe1
		0x0101, /* bcdDevice */        // 1.1
		0x01, /* iManufacturer */      // Manufacturer
		0x02, /* iProduct */           // Product
		0x03, /* iSerialNumber */      // Serial Number
		0x01, /* bNumConfigurations */
};

/**
 @name config_descriptor_cdcacm
 @brief Configuration descriptor for Run Time mode.
 */
//@{
struct PACK config_descriptor_cdcacm
{
	USB_configuration_descriptor configuration;
	USB_interface_association_descriptor interface_association;
	USB_interface_descriptor interface_control;
	USB_CDC_ClassSpecificDescriptorHeaderFormat cdcCPDHF;
	USB_CDC_UnionInterfaceFunctionalDescriptor cdcUIFD;
	USB_CDC_CallManagementFunctionalDescriptor cdcCMFD;
	USB_CDC_AbstractControlManagementFunctionalDescriptor cdcACMFD;
	USB_endpoint_descriptor endpoint_control;
	USB_interface_descriptor interface_data;
	USB_endpoint_descriptor endpoint_data_out;
	USB_endpoint_descriptor endpoint_data_in;
	USB_interface_descriptor dfu_interface;
	USB_dfu_functional_descriptor dfu_functional;
};

// Fields marked with * are updated with information in device_config[]
DESCRIPTOR_QUALIFIER struct config_descriptor_cdcacm config_descriptor_cdcacm =
{
	{
		sizeof(USB_configuration_descriptor), /* configuration.bLength */
		USB_DESCRIPTOR_TYPE_CONFIGURATION, /* configuration.bDescriptorType */
		sizeof(struct config_descriptor_cdcacm), /* configuration.wTotalLength */
		0x03, /* configuration.bNumInterfaces */
		0x01, /* configuration.bConfigurationValue */
		0x00, /* configuration.iConfiguration */
		 USB_CONFIG_BMATTRIBUTES_VALUE, /* configuration.bmAttributes */
		0xFA /* configuration.bMaxPower */           // 500mA
	},

	{
		sizeof(USB_interface_association_descriptor), /* interface_association.bLength */
		USB_DESCRIPTOR_TYPE_INTERFACE_ASSOCIATION, /* interface_association.bDescriptorType */
		0, /* interface_association.bFirstInterface */
		2, /* interface_association.bInterfaceCount */
		USB_CLASS_CDC_CONTROL, /* interface_association.bFunctionClass */
		USB_SUBCLASS_CDC_CONTROL_ABSTRACT, /* interface_association.bFunctionSubClass */
		USB_PROTOCOL_CDC_CONTROL_ITU_T_V250, /* interface_association.bFunctionProtocol */
		2 /* interface_association.iFunction */  // "FT900 CDCACM"
	},

	// ---- INTERFACE DESCRIPTOR for CDCACM Control ----
	{
		sizeof(USB_interface_descriptor), /* interface_control.bLength */
		USB_DESCRIPTOR_TYPE_INTERFACE, /* interface_control.bDescriptorType */
		0, /* interface_control.bInterfaceNumber */
		0x00, /* interface_control.bAlternateSetting */
		0x01, /* interface_control.bNumEndpoints */
		USB_CLASS_CDC_CONTROL, /* interface_control.bInterfaceClass */ // CDC Class
		USB_SUBCLASS_CDC_CONTROL_ABSTRACT, /* interface_control.bInterfaceSubClass */ // Abstract Control Model
		USB_PROTOCOL_CDC_CONTROL_ITU_T_V250, /* interface_control.bInterfaceProtocol */ // No built-in protocol
		0x00 /* interface_control.iInterface */ // Unused
	},

	// ---- FUNCTIONAL DESCRIPTORS for CDCACM ----
	{
		sizeof(USB_CDC_ClassSpecificDescriptorHeaderFormat), /* cdcCPDHF.bFunctionLength */
		USB_CDC_DESCRIPTOR_TYPE_CS_INTERFACE, /* cdcCPDHF.bDescriptorType */
		USB_CDC_DESCRIPTOR_SUBTYPE_HEADER_FUNCTIONAL_DESCRIPTOR, /* cdcCPDHF.bDescriptorSubtype */
		0x0110 /* cdcCPDHF.bcdCDC */
	},

	{
		sizeof(USB_CDC_UnionInterfaceFunctionalDescriptor), /* cdcUIFD.bFunctionLength */
		USB_CDC_DESCRIPTOR_TYPE_CS_INTERFACE, /* cdcUIFD.bDescriptorType */
		USB_CDC_DESCRIPTOR_SUBTYPE_UNION_FUNCTIONAL_DESCRIPTOR, /* cdcUIFD.bDescriptorSubtype */
		0, /* cdcUIFD.bControlInterface */
		1 /* cdcUIFD.bSubordinateInterface0 */
	},

	{
		sizeof(USB_CDC_CallManagementFunctionalDescriptor), /* cdcCMFD.bFunctionLength */
		USB_CDC_DESCRIPTOR_TYPE_CS_INTERFACE, /* cdcCMFD.bDescriptorType */
		USB_CDC_DESCRIPTOR_SUBTYPE_CALL_MANAGEMENT_FUNCTIONAL_DESCRIPTOR, /* cdcCMFD.bDescriptorSubtype */
		USB_CDC_CM_CAPABILITIES_NONE, /* cdcCMFD.bmCapabilities */
		0x01 /* cdcCMFD.bDataInterface */
	},

	{
		sizeof(USB_CDC_AbstractControlManagementFunctionalDescriptor), /* cdcACMFD.bFunctionLength */
		USB_CDC_DESCRIPTOR_TYPE_CS_INTERFACE, /* cdcACMFD.bDescriptorType */
		USB_CDC_DESCRIPTOR_SUBTYPE_ABSTRACT_CONTROL_MANAGEMENT_FUNCTIONAL_DESCRIPTOR, /* cdcACMFD.bDescriptorSubtype */
		USB_CDC_ACM_CAPABILITIES_LINE_STATE_CODING /* cdcACMFD.bmCapabilities */
	},

	// ---- ENDPOINT DESCRIPTOR for CDCACM Control ----
	{
		sizeof(USB_endpoint_descriptor), /* endpoint_control.bLength */
		USB_DESCRIPTOR_TYPE_ENDPOINT, /* endpoint_control.bDescriptorType */
		USB_ENDPOINT_DESCRIPTOR_EPADDR_IN | CDC_EP_NOTIFICATION, /* endpoint_control.bEndpointAddress */
		USB_ENDPOINT_DESCRIPTOR_ATTR_INTERRUPT, /* endpoint_control.bmAttributes */
		CDC_NOTIFICATION_EP_SIZE, /* endpoint_control.wMaxPacketSize */
		0x0a /* endpoint_control.bInterval */
	},

	// ---- INTERFACE DESCRIPTOR for CDCACM Data ----
	{
		sizeof(USB_interface_descriptor), /* interface_data.bLength */
		USB_DESCRIPTOR_TYPE_INTERFACE, /* interface_data.bDescriptorType */
		1, /* interface_data.bInterfaceNumber */
		0x00, /* interface_data.bAlternateSetting */
		0x02, /* interface_data.bNumEndpoints */
		USB_CLASS_CDC_DATA, /* interface_data.bInterfaceClass */ // CDC Data Class
		USB_SUBCLASS_CDC_DATA, /* interface_data.bInterfaceSubClass */ // Abstract Control Model
		USB_PROTOCOL_CDC_DATA, /* interface_data.bInterfaceProtocol */ // No built-in protocol
		0x00 /* interface_data.iInterface */               // Unused
	},

	// ---- ENDPOINT DESCRIPTORS for CDCACM Data ----
	{
		sizeof(USB_endpoint_descriptor), /* endpoint_data_out.bLength */
		USB_DESCRIPTOR_TYPE_ENDPOINT, /* endpoint_data_out.bDescriptorType */
		USB_ENDPOINT_DESCRIPTOR_EPADDR_OUT | CDC_EP_DATA_OUT, /* endpoint_data_out.bEndpointAddress */
		USB_ENDPOINT_DESCRIPTOR_ATTR_BULK, /* endpoint_data_out.bmAttributes */
		CDC_DATA_EP_SIZE, /* endpoint_data_out.wMaxPacketSize */
		0x0a /* endpoint_data_out.bInterval */
	},

	{
		sizeof(USB_endpoint_descriptor), /* endpoint_data_in.bLength */
		USB_DESCRIPTOR_TYPE_ENDPOINT, /* endpoint_data_in.bDescriptorType */
		USB_ENDPOINT_DESCRIPTOR_EPADDR_IN | CDC_EP_DATA_IN, /* endpoint_data_in.bEndpointAddress */
		USB_ENDPOINT_DESCRIPTOR_ATTR_BULK, /* endpoint_data_in.bmAttributes */
		CDC_DATA_EP_SIZE, /* endpoint_data_in.wMaxPacketSize */
		0x0a /* endpoint_data_in.bInterval */
	},

	// ---- INTERFACE DESCRIPTOR for DFU Interface ----
	{
		sizeof(USB_interface_descriptor), /* dfu_interface.bLength */
		USB_DESCRIPTOR_TYPE_INTERFACE, /* dfu_interface.bDescriptorType */
		DFU_USB_INTERFACE_RUNTIME, /* dfu_interface.bInterfaceNumber */
		0x00, /* dfu_interface.bAlternateSetting */
		0x00, /* dfu_interface.bNumEndpoints */
		USB_CLASS_APPLICATION, /* dfu_interface.bInterfaceClass */ // bInterfaceClass: Application Specific Class
		USB_SUBCLASS_DFU, /* dfu_interface.bInterfaceSubClass */ // bInterfaceSubClass: Device Firmware Update
		USB_PROTOCOL_DFU_RUNTIME, /* dfu_interface.bInterfaceProtocol */ // bInterfaceProtocol: Runtime Protocol
		0x05 /* dfu_interface.iInterface */       // * iInterface: "DFU Interface"
	},

	// ---- FUNCTIONAL DESCRIPTOR for DFU Interface ----
	{
		sizeof(USB_dfu_functional_descriptor), /* dfu_functional.bLength */
		USB_DESCRIPTOR_TYPE_DFU_FUNCTIONAL, /* dfu_functional.bDescriptorType */
		DFU_ATTRIBUTES, /* dfu_functional.bmAttributes */  	// bmAttributes
		DFU_TIMEOUT, /* dfu_functional.wDetatchTimeOut */ // wDetatchTimeOut
		DFU_TRANSFER_SIZE, /* dfu_functional.wTransferSize */     // wTransferSize
		USB_BCD_VERSION_DFU_1_1 /* dfu_functional.bcdDfuVersion */ // bcdDfuVersion: DFU Version 1.1
	}
};
//@}

/**
 @name wcid_feature_runtime
 @brief WCID Compatible ID for DFU interface in runtime.
 */
//@{
DESCRIPTOR_QUALIFIER USB_WCID_feature_descriptor wcid_feature_runtime =
{
	sizeof(struct USB_WCID_feature_descriptor), /* dwLength */
	USB_MICROSOFT_WCID_VERSION, /* bcdVersion */
	USB_MICROSOFT_WCID_FEATURE_WINDEX_COMPAT_ID, /* wIndex */
	1, /* bCount */
	{0, 0, 0, 0, 0, 0, 0,}, /* rsv1 */
	DFU_USB_INTERFACE_RUNTIME, /* bFirstInterfaceNumber */
	1, /* rsv2 - set to 1 */
	{'W', 'I', 'N', 'U', 'S', 'B', 0x00, 0x00,}, /* compatibleID[8] */
	{0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,}, /* subCompatibleID[8] */
	{0, 0, 0, 0, 0, 0,}, /* rsv3[6] */
};

/**
 @name device_descriptor_dfumode
 @brief Device descriptor for DFU Mode.
 */
DESCRIPTOR_QUALIFIER USB_device_descriptor device_descriptor_dfumode =
{
		0x12, /* bLength */
		USB_DESCRIPTOR_TYPE_DEVICE, /* bDescriptorType */
		USB_BCD_VERSION_2_0, /* bcdUSB */          // V2.00
		USB_CLASS_DEVICE, /* bDeviceClass */       // Defined in interface
		USB_SUBCLASS_DEVICE, /* bDeviceSubClass */ // Defined in interface
		USB_PROTOCOL_DEVICE, /* bDeviceProtocol */ // Defined in interface
		USB_CONTROL_EP_MAX_PACKET_SIZE, /* bMaxPacketSize0 */    // 8
		USB_VID_FTDI, /* idVendor */   // idVendor: 0x0403 (FTDI)
		DFU_USB_PID_DFUMODE, /* idProduct */ // idProduct: 0x0fee
		0x0101, /* bcdDevice */        // 1.1
		0x01, /* iManufacturer */      // Manufacturer
		0x04, /* iProduct */           // Product
		0x03, /* iSerialNumber */      // Serial Number
		0x01, /* bNumConfigurations */
};

/**
 @name config_descriptor_dfumode
 @brief Config descriptor for DFU Mode.
 */
//@{
struct PACK config_descriptor_dfumode
{
	USB_configuration_descriptor configuration;
	USB_interface_descriptor dfu_interface;
	USB_dfu_functional_descriptor dfu_functional;
};

DESCRIPTOR_QUALIFIER struct config_descriptor_dfumode config_descriptor_dfumode =
{
	{
		0x09, /* configuration.bLength */
		USB_DESCRIPTOR_TYPE_CONFIGURATION, /* configuration.bDescriptorType */
		sizeof(struct config_descriptor_dfumode), /* configuration.wTotalLength */
		0x01, /* configuration.bNumInterfaces */
		0x01, /* configuration.bConfigurationValue */
		0x00, /* configuration.iConfiguration */
		USB_CONFIG_BMATTRIBUTES_VALUE, /* configuration.bmAttributes */
		0xFA /* configuration.bMaxPower */ // 500mA
	},

	// ---- INTERFACE DESCRIPTOR for DFU Interface ----
	{
		0x09, /* dfu_interface.bLength */
		USB_DESCRIPTOR_TYPE_INTERFACE, /* dfu_interface.bDescriptorType */
		DFU_USB_INTERFACE_DFUMODE, /* dfu_interface.bInterfaceNumber */
		0x00, /* dfu_interface.bAlternateSetting */
		0x00, /* dfu_interface.bNumEndpoints */
		USB_CLASS_APPLICATION, /* dfu_interface.bInterfaceClass */ // bInterfaceClass: Application Specific Class
		USB_SUBCLASS_DFU, /* dfu_interface.bInterfaceSubClass */ // bInterfaceSubClass: Device Firmware Update
		USB_PROTOCOL_DFU_DFUMODE, /* dfu_interface.bInterfaceProtocol */ // bInterfaceProtocol: Runtime Protocol
		0x05 /* dfu_interface.iInterface */
	},

	// ---- FUNCTIONAL DESCRIPTOR for DFU Interface ----
	{
		0x09, /* dfu_functional.bLength */
		USB_DESCRIPTOR_TYPE_DFU_FUNCTIONAL, /* dfu_functional.bDescriptorType */
		DFU_ATTRIBUTES, /* dfu_functional.bmAttributes */  	// bmAttributes
		DFU_TIMEOUT, /* dfu_functional.wDetatchTimeOut */ // wDetatchTimeOut
		DFU_TRANSFER_SIZE, /* dfu_functional.wTransferSize */     // wTransferSize
		USB_BCD_VERSION_DFU_1_1 /* dfu_functional.bcdDfuVersion */ // bcdDfuVersion: DFU Version 1.1
	}
};
//@}

/**
 @name wcid_feature_dfumode
 @brief WCID Compatible ID for DFU interface in DFU mode.
 */
//@{
DESCRIPTOR_QUALIFIER USB_WCID_feature_descriptor wcid_feature_dfumode =
{
	sizeof(struct USB_WCID_feature_descriptor), /* dwLength */
	USB_MICROSOFT_WCID_VERSION, /* bcdVersion */
	USB_MICROSOFT_WCID_FEATURE_WINDEX_COMPAT_ID, /* wIndex */
	1, /* bCount */
	{0, 0, 0, 0, 0, 0, 0,}, /* rsv1 */
	DFU_USB_INTERFACE_DFUMODE, /* bFirstInterfaceNumber */
	1, /* rsv2 - set to 1 */
	{'W', 'I', 'N', 'U', 'S', 'B', 0x00, 0x00,}, /* compatibleID[8] */
	{0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,}, /* subCompatibleID[8] */
	{0, 0, 0, 0, 0, 0,}, /* rsv3[6] */
};

/**
 @name USBD_cdc_line_coding
 @brief UART characteristics set by CDC ACM driver on host.
 */
static USB_CDC_line_coding cdc_line_coding =
{
		19200, /* dwDTERate */
		USB_CDC_STOP_BITS_1, /* bCharFormat */
		USB_CDC_PARITY_NONE, /* bParityType */
		USB_CDC_DATA_BITS_8, /* bDataBits */
};


/**
 @name UART ring buffer variables.
 @brief Variables required to implement an Rx and Tx ring buffer for the UART.
 */
//@{
static uint8_t uart0BufferIn_data[RINGBUFFER_SIZE];
static volatile uint16_t uart0BufferIn_wr_idx = 0;
static volatile uint16_t uart0BufferIn_rd_idx = 0;
static volatile uint16_t uart0BufferIn_avail = RINGBUFFER_SIZE;
static uint8_t uart0BufferOut_data[RINGBUFFER_SIZE];
static volatile uint16_t uart0BufferOut_wr_idx = 0;
static volatile uint16_t uart0BufferOut_rd_idx = 0;
static volatile uint16_t uart0BufferOut_avail = RINGBUFFER_SIZE;
//@}

/* MACROS **************************************************************************/

/* LOCAL FUNCTIONS / INLINES *******************************************************/

size_t uart0Tx(uint8_t *data, size_t len);
size_t uart0Rx(uint8_t *data, size_t len);
void uart0ISR(void);

/** @name tfp_putc
 *  @details Machine dependent putc function for tfp_printf (tinyprintf) library.
 *  @param p Parameters (machine dependent)
 *  @param c The character to write
 */
void tfp_putc(void* p, char c)
{
	uart_write((ft900_uart_regs_t*)p, (uint8_t)c);
}

void ISR_timer(void)
{
	if (timer_is_interrupted(timer_select_a))
	{
		// Call USB timer handler to action transaction and hub timers.
		USBD_timer();

		// Call the DFU timer to progress state machine.
		// The progression depends on the bmAttributes set in the configuration.
		USBD_DFU_timer();
	}
}

/**
 Transmit a buffer of data over UART0 asynchronously.

 @return The number of bytes written to the buffer
 */
size_t uart0Tx(uint8_t *buffer, size_t len)
{
	signed int start;
	size_t copied = 0;

	// Determine if we need to start a transmission or there is one
	// already in action.
	CRITICAL_SECTION_BEGIN
	start = uart0BufferOut_wr_idx - uart0BufferOut_rd_idx;
	CRITICAL_SECTION_END

	CRITICAL_SECTION_BEGIN
	// Copy in as much data from the producer as there is space for
	// in the out buffer.
	while (len--)
	{
		// Check there is space there to receive the data from the producer.
		if (uart0BufferOut_avail > 0)
		{
			// Reduce the number of available bytes.
			uart0BufferOut_avail--;

			// Add byte of data to the buffer.
			uart0BufferOut_data[uart0BufferOut_wr_idx] = *buffer++;

			// Increment the pointer and wrap around.
			uart0BufferOut_wr_idx++;
			uart0BufferOut_wr_idx &= (RINGBUFFER_SIZE - 1);

			// Tally up number of bytes actually transmitted.
			copied++;
		}
	}
	CRITICAL_SECTION_END

	// Start new a transmission if nothing is being transmitted.
	if (start == 0)
	{
		uint8_t c;

		CRITICAL_SECTION_BEGIN

		// Get data from the out buffer.
		c = uart0BufferOut_data[uart0BufferOut_rd_idx];

		// Increment the consumer pointer and wrap around.
		uart0BufferOut_rd_idx++;
		uart0BufferOut_rd_idx &= (RINGBUFFER_SIZE - 1);

		// Byte has been sent, update available bytes value.
		uart0BufferOut_avail++;

		CRITICAL_SECTION_END

		// Send data to the UART. This will result in a TX interrupt when
		// it completes.
		uart_write(UART0, c);
	}

	return copied;
}

/**
 Receive a number of bytes from UART0

 @return The number of bytes read from UART0
 */
size_t uart0Rx(uint8_t *buffer, size_t len)
{
	size_t copied = 0;

	// Copy in as much data as we can ...
	//   This can be either the maximum size of the buffer being given
	//   or the maximum number of bytes available in the Serial Port
	//   buffer.
	while(len--)
	{
		CRITICAL_SECTION_BEGIN
		if (uart0BufferIn_avail < RINGBUFFER_SIZE)
		{
			uart0BufferIn_avail++;

			*buffer = uart0BufferIn_data[uart0BufferIn_rd_idx];
			buffer++;

			// Increment the pointer and wrap around.
			uart0BufferIn_rd_idx++;
			uart0BufferIn_rd_idx &= (RINGBUFFER_SIZE - 1);

			copied++;
		}
		CRITICAL_SECTION_END
	}

	// Report back how many bytes have been copied into the buffer...
	return copied;
}

/**
 The Interrupt which handles asynchronous transmission and reception
 of data into the ring buffer
 */
void uart0ISR()
{
	static uint8_t c;

	// Transmit interrupt...
	if (uart_is_interrupted(UART0, uart_interrupt_tx))
	{
		// Check to see how much data we have to transmit...
		//avail = uart0BufferOut_rd_idx - uart0BufferOut_wr_idx;

		if (uart0BufferOut_avail < RINGBUFFER_SIZE)
		{
			uart0BufferOut_avail++;

			// Copy out the byte to be transmitted so that the uart_write is
			// the last thing we do...
			c = uart0BufferOut_data[uart0BufferOut_rd_idx];

			// Increment the pointer and wrap around.
			uart0BufferOut_rd_idx++;
			uart0BufferOut_rd_idx &= (RINGBUFFER_SIZE - 1);

			// Write out a new byte, the following Transmit interrupt should handle
			// the remaning bytes...
			uart_write(UART0, c);
		}
	}

	// Receive interrupt...
	if (uart_is_interrupted(UART0, uart_interrupt_rx))
	{
		// Read a byte into the Ring Buffer...
		uart_read(UART0, &c);

		// Do not overwrite data already in the buffer.
		if (uart0BufferIn_avail > 0)
		{
			uart0BufferIn_avail--;

			uart0BufferIn_data[uart0BufferIn_wr_idx] = c;

			// Increment the pointer and wrap around.
			uart0BufferIn_wr_idx++;
			uart0BufferIn_wr_idx &= (RINGBUFFER_SIZE - 1);
		}
	}
}

void uartCoding(void)
{
	uint16_t divisor;
	uint8_t prescaler;
	uint8_t dataBits;
	uint8_t parity;
	uint8_t stop;

	uart_calculate_baud(cdc_line_coding.dwDTERate, 4, 100000000, &divisor, &prescaler);

	switch(cdc_line_coding.bDataBits)
	{
	// Data bites (5,6,7,8 or 16) NB 16 bits not supported
	default:
	case 8: dataBits = uart_data_bits_8; break;
	case 7: dataBits = uart_data_bits_7; break;
	case 6: dataBits = uart_data_bits_6; break;
	case 5: dataBits = uart_data_bits_5; break;
	};
	switch(cdc_line_coding.bParityType)
	{
	// Parity: 0 - None; 1 - Odd; 2 - Even; 3 - Mark; 4 - Space
	// NB Mark and Space not supported.
	default:
	case 0: parity = uart_parity_none; break;
	case 1: parity = uart_parity_odd; break;
	case 2: parity = uart_parity_even; break;
	};
	switch(cdc_line_coding.bCharFormat)
	{
	// Stop bits: 0 - 1 Stop bit; 1 - 1.5 Stop bit; 2 - 2 Stop bits
	default:
	case 0: stop = uart_stop_bits_1; break;
	case 1: stop = uart_stop_bits_1_5; break;
	case 2: stop = uart_stop_bits_2; break;
	};

	uart_open(UART0, prescaler, divisor, dataBits, parity, stop);
}

/**
 @brief      USB class request handler
 @details    Handle class requests from the host application.
 The bRequest value is parsed and the appropriate
 action or function is performed. Additional values
 from the USB_device_request structure are decoded
 and the DFU state machine and status updated as
 required.
 @param[in]	req - USB_device_request structure containing the
 SETUP portion of the request from the host.
 @return		status - USBD_OK if successful or USBD_ERR_*
 if there is an error or the bRequest is not
 supported.
 **/
int8_t class_req_cb(USB_device_request *req)
{
	int8_t status = USBD_ERR_NOT_SUPPORTED;
	uint8_t interface = LSB(req->wIndex) & 0x0F;
	uint8_t command[USB_CONTROL_EP_MAX_PACKET_SIZE];
	uint8_t request = req->bRequest;

	// For DFU requests the SETUP packet must be addressed to the
	// the DFU interface on the device.

	// For DFU requests ensure the recipient is an interface...
	if ((req->bmRequestType & USB_BMREQUESTTYPE_RECIPIENT_MASK) ==
			USB_BMREQUESTTYPE_RECIPIENT_INTERFACE)
	{
		// ...and that the interface is the correct Runtime interface
		if (USBD_DFU_is_runtime())
		{
			if ((interface == DFU_USB_INTERFACE_RUNTIME))
			{
				// Handle only DFU_DETATCH, DFU_GETSTATE and DFU_GETSTATUS
				// when in Runtime mode. Table 3.2 DFU_DETACH is mandatory
				// in Runtime mode, DFU_GETSTATE and DFU_GETSTATUS are
				// optional.
				switch (request)
				{
				case USB_CLASS_REQUEST_DETACH:
					USBD_DFU_class_req_detach(req->wValue);
					status = USBD_OK;
					break;
				case USB_CLASS_REQUEST_GETSTATUS:
					USBD_DFU_class_req_getstatus(req->wLength);
					status = USBD_OK;
					break;
				case USB_CLASS_REQUEST_GETSTATE:
					USBD_DFU_class_req_getstate(req->wLength);
					status = USBD_OK;
					break;
				}
			}
			else
			{
				switch (request)
				{
				case USB_CDC_REQUEST_SEND_ENCAPSULATED_COMMAND:
					// Read and ignore encapsulated command.
					USBD_transfer_ep0(USBD_DIR_OUT, command, USB_CONTROL_EP_MAX_PACKET_SIZE, req->wLength);
					// ACK
					USBD_transfer_ep0(USBD_DIR_IN, NULL, 0, 0);
					break;
				case USB_CDC_REQUEST_GET_ENCAPSULATED_RESPONSE:
				case USB_CDC_REQUEST_SET_COMM_FEATURE:
				case USB_CDC_REQUEST_GET_COMM_FEATURE:
				case USB_CDC_REQUEST_CLEAR_COMM_FEATURE:
					break;

				case USB_CDC_REQUEST_SET_LINE_CODING:
					USBD_transfer_ep0(USBD_DIR_OUT, (uint8_t*)&cdc_line_coding, sizeof(USB_CDC_line_coding), req->wLength);
					// ACK
					USBD_transfer_ep0(USBD_DIR_IN, NULL, 0, 0);
					// Update UART settings
					tfp_printf("set baud %ld, par %d, stop %d, data %d\r\n",
							cdc_line_coding.dwDTERate,
							cdc_line_coding.bParityType,
							cdc_line_coding.bCharFormat,
							cdc_line_coding.bDataBits);
					uartCoding();
					status = USBD_OK;
					break;

				case USB_CDC_REQUEST_GET_LINE_CODING:
					USBD_transfer_ep0(USBD_DIR_IN, (uint8_t*)&cdc_line_coding, sizeof(USB_CDC_line_coding), req->wLength);
					// ACK
					USBD_transfer_ep0(USBD_DIR_OUT, NULL, 0, 0);
					status = USBD_OK;
					tfp_printf("get baud %ld, par %d, stop %d, data %d\r\n",
							cdc_line_coding.dwDTERate,
							cdc_line_coding.bParityType,
							cdc_line_coding.bCharFormat,
							cdc_line_coding.bDataBits);
					break;

				case USB_CDC_REQUEST_SET_CONTROL_LINE_STATE:
					// Ignore the control line state
					// ACK
					USBD_transfer_ep0(USBD_DIR_IN, NULL, 0, 0);
					status = USBD_OK;
					break;

				case USB_CDC_REQUEST_SEND_BREAK :
					//USBD_transfer_ep0(USBD_DIR_IN, NULL, 0, 0);
					break;

				}
			}
		}
		// ...or the correct DFU Mode interface
		else
		{
			if (interface == DFU_USB_INTERFACE_DFUMODE)
			{
				// Handle remaining DFU class requests when in DFU Mode.
				// Table 3.2 DFU_DETACH is not supported in DFU Mode.
				switch (req->bRequest)
				{
				case USB_CLASS_REQUEST_DNLOAD:
					/* Block number passed in wValue gives the start address of
					 * to program based on the size of the transfer size.
					 */
					USBD_DFU_class_req_download(req->wValue * DFU_TRANSFER_SIZE,
							req->wLength);
					status = USBD_OK;
					break;

				case USB_CLASS_REQUEST_UPLOAD:
					/* Block number passed in wValue gives the start address of
					 * to program based on the size of the transfer size.
					 */
					USBD_DFU_class_req_upload(req->wValue * DFU_TRANSFER_SIZE,
							req->wLength);
					status = USBD_OK;
					break;

				case USB_CLASS_REQUEST_GETSTATUS:
					USBD_DFU_class_req_getstatus(req->wLength);
					status = USBD_OK;
					break;

				case USB_CLASS_REQUEST_GETSTATE:
					USBD_DFU_class_req_getstate(req->wLength);
					status = USBD_OK;
					break;
				case USB_CLASS_REQUEST_CLRSTATUS:
					USBD_DFU_class_req_clrstatus();
					status = USBD_OK;
					break;
				case USB_CLASS_REQUEST_ABORT:
					USBD_DFU_class_req_abort();
					status = USBD_OK;
					break;

				default:
					// Unknown or unsupported request.
					break;
				}
			}
		}
	}

	return status;
}

/**
 @brief      USB standard request handler for GET_DESCRIPTOR
 @details    Handle standard GET_DESCRIPTOR requests from the host
 application.
 The hValue is parsed and the appropriate device,
 configuration or string descriptor is selected.
 For device and configuration descriptors the DFU
 state machine determines which descriptor to use.
 For string descriptors the lValue selects which
 string from the string_descriptors table to use.
 The string table is automatically traversed to find
 the correct string and the length is taken from the
 string descriptor.
 @param[in]	req - USB_device_request structure containing the
 SETUP portion of the request from the host.
 @return		status - USBD_OK if successful or USBD_ERR_*
 if there is an error or the bmRequestType is not
 supported.
 **/
int8_t standard_req_get_descriptor(USB_device_request *req, uint8_t **buffer, uint16_t *len)
{
	uint8_t *src = NULL;
	uint16_t length = req->wLength;
	uint8_t hValue = req->wValue >> 8;
	uint8_t lValue = req->wValue & 0x00ff;
	uint8_t i;
	uint8_t slen;

	switch (hValue)
	{
	case USB_DESCRIPTOR_TYPE_DEVICE:
		if (USBD_DFU_is_runtime())
		{
			src = (uint8_t *) &device_descriptor_cdcacm;
		}
		else
		{
			src = (uint8_t *) &device_descriptor_dfumode;
		}
		if (length > sizeof(USB_device_descriptor)) // too many bytes requested
				length = sizeof(USB_device_descriptor); // Entire structure.
		break;

	case USB_DESCRIPTOR_TYPE_CONFIGURATION:
		if (USBD_DFU_is_runtime())
		{
			src = (uint8_t *) &config_descriptor_cdcacm;
			if (length > sizeof(config_descriptor_cdcacm)) // too many bytes requested
				length = sizeof(config_descriptor_cdcacm); // Entire structure.
		}
		else
		{
			src = (uint8_t *) &config_descriptor_dfumode;
			if (length > sizeof(config_descriptor_dfumode)) // too many bytes requested
				length = sizeof(config_descriptor_dfumode); // Entire structure.
		}
		break;

	case USB_DESCRIPTOR_TYPE_STRING:
		// Special case: WCID descriptor
		if (lValue == USB_MICROSOFT_WCID_STRING_DESCRIPTOR)
		{
			src = (uint8_t *)wcid_string_descriptor;
			length = sizeof(wcid_string_descriptor);
			break;
		}

		// Find the nth string in the string descriptor table
		i = 0;
		while ((slen = string_descriptor[i]) > 0)
		{
			// Point to start of string descriptor in __code segment
			// Typecast prevents warning for losing const when USBD_transfer
			// is called
			src = (uint8_t *) &string_descriptor[i];
			if (lValue == 0)
			{
				break;
			}
			i += slen;
			lValue--;
		}
		if (lValue > 0)
		{
			// String not found
			return USBD_ERR_NOT_SUPPORTED;
		}
		// Update the length returned only if it is less than the requested
		// size
		if (length > slen)
		{
			length = slen;
		}

		break;

	default:
		return USBD_ERR_NOT_SUPPORTED;
	}

	*buffer = src;
	*len = length;

	return USBD_OK;
}

/**
 @brief      USB vendor request handler
 @details    Handle vendor requests from the host application.
 The bRequest value is parsed and the appropriate
 action or function is performed. Additional values
 from the USB_device_request structure are decoded
 and provided to other handlers.
 There are no vendor requests requiring handling in
 this firmware.
 @param[in]	req - USB_device_request structure containing the
 SETUP portion of the request from the host.
 @return		status - USBD_OK if successful or USBD_ERR_*
 if there is an error or the bRequest is not
 supported.
 **/
int8_t vendor_req_cb(USB_device_request *req)
{
	int8_t status = USBD_ERR_NOT_SUPPORTED;
	uint16_t length = req->wLength;

	// For Microsoft WCID only.
	// Request for Compatible ID Feature Descriptors.
	// Check the request code and wIndex.
	if (req->bRequest == WCID_VENDOR_REQUEST_CODE)
	{
		if (req->wIndex == USB_MICROSOFT_WCID_FEATURE_WINDEX_COMPAT_ID)
		{
			if (req->bmRequestType & USB_BMREQUESTTYPE_DIR_DEV_TO_HOST)
			{
				if (length > sizeof(wcid_feature_runtime)) // too many bytes requested
					length = sizeof(wcid_feature_runtime); // Entire structure.
				// Return a compatible ID feature descriptor.
				if (USBD_DFU_is_runtime())
				{
					USBD_transfer_ep0(USBD_DIR_IN, (uint8_t *) &wcid_feature_runtime,
							length, length);
				}
				else
				{
					USBD_transfer_ep0(USBD_DIR_IN, (uint8_t *) &wcid_feature_dfumode,
							length, length);
				}
				// ACK packet
				USBD_transfer_ep0(USBD_DIR_OUT, NULL, 0, 0);
				status = USBD_OK;
			}
		}
	}

	return status;
}

/**
 @brief      USB suspend callback
 @details    Called when the USB bus enters the suspend state.
 @param[in]	status - unused.
 @return		N/A
 **/
void suspend_cb(uint8_t status)
{
	(void) status;
	return;
}

/**
 @brief      USB resume callback
 @details    Called when the USB bus enters the resume state
 prior to restating after a suspend.
 @param[in]  status - unused.
 @return     N/S
 **/
void resume_cb(uint8_t status)
{
	(void) status;
	return;
}

/**
 @brief      USB reset callback
 @details    Called when the USB bus enters the reset state.
 @param[in]	status - unused.
 @return		N/A
 **/
void reset_cb(uint8_t status)
{
	(void) status;

	USBD_set_state(USBD_STATE_DEFAULT);

	// If we are at DFUSTATE_MANIFEST_WAIT_RESET stage and do
	// not support detach then the DFU will reset the chip and
	// run the new firmware.
	// From the DFUSTATE_APPIDLE state advance to DFUSTATE_DFUIDLE.
	// Other states will not advance the state machine or may
	// move this to DFUSTATE_DFUERROR if it is inappropriate to
	// find a reset at that stage.
	USBD_DFU_reset();

	return;
}

/* For debugging endpoint transactions. */
void ep_cb(USBD_ENDPOINT_NUMBER ep_number)
{
	tfp_printf("EP%d\r\n", ep_number);
}

void cdc_send_serial_state_notification()
{
	struct { // Note: structure must be packed
		USB_CDC_Notification header;
		union {
			uint16_t buffer;
			USB_CDC_UartStateBitmap serialState;
		};
	} notification;

	notification.header.bmRequestType = USB_BMREQUESTTYPE_DIR_DEV_TO_HOST |
			USB_BMREQUESTTYPE_CLASS |
			USB_BMREQUESTTYPE_RECIPIENT_INTERFACE;
	notification.header.bNotification = USB_CDC_NOTIFICAITON_SERIAL_STATE;
	notification.header.wValue = 0;
	notification.header.wIndex = 0; // CDC ACM interface
	notification.header.wLength = 2;
	notification.buffer = 0;
	// All Serial State bits are set to zero (disabled).
	notification.serialState.bRxCarrier = 0;
	notification.serialState.bTxCarrier = 0;

	USBD_transfer(CDC_EP_NOTIFICATION, (uint8_t *) &notification, sizeof(notification));
}

void cdc_uart_bridge(void)
{
	int8_t status;
	static uint8_t bufUart2Usb[CDC_DATA_EP_SIZE] = {0};
	static uint8_t bufUsb2Uart[CDC_DATA_EP_SIZE] = {0};
	size_t read_bytes;

	USBD_create_endpoint(CDC_EP_NOTIFICATION, USBD_EP_INT, USBD_DIR_IN,
			CDC_NOTIFICATION_USBD_EP_SIZE, USBD_DB_OFF, NULL /*ep_cb*/);
	USBD_create_endpoint(CDC_EP_DATA_OUT, USBD_EP_BULK, USBD_DIR_OUT,
			CDC_DATA_USBD_EP_SIZE, USBD_DB_ON, NULL /*ep_cb*/);
	USBD_create_endpoint(CDC_EP_DATA_IN, USBD_EP_BULK, USBD_DIR_IN,
			CDC_DATA_USBD_EP_SIZE, USBD_DB_ON, NULL /*ep_cb*/);

	cdc_send_serial_state_notification();

	do
	{
		// If the USB IN endpoint is empty then write data from the UART to
		// the host. The data in the bufUart2Usb has been moved into the
		// endpoint data RAM by USBD_transfer.
		if (!USBD_ep_buffer_full(CDC_EP_DATA_IN))
		{
			// Read in a packet of data from the UART.
			// (up to the size of the ring buffer)
			read_bytes = uart0Rx(bufUart2Usb, CDC_DATA_EP_SIZE);

			if (read_bytes)
			{
				// Write it to the CDC DATA interface.
				USBD_transfer(CDC_EP_DATA_IN, (uint8_t *) &bufUart2Usb, read_bytes);
			}
		}

		// If data is available on the USB OUT endpoint and there is space
		// in the ring buffer to receive it then read it in from the host.
		if ((USBD_ep_buffer_full(CDC_EP_DATA_OUT)) &&
				(uart0BufferOut_avail >= CDC_DATA_EP_SIZE))
		{
			// Read in a packet of data from the CDC DATA interface.
			read_bytes = USBD_transfer(CDC_EP_DATA_OUT, (uint8_t *) &bufUsb2Uart, CDC_DATA_EP_SIZE);
			if (read_bytes)
			{
				// Write it out to the UART.
				uart0Tx(bufUsb2Uart, read_bytes);
			}
		}

		// Update USBH status and perform callbacks if required.
		status = USBD_process();
	}
	while (status == USBD_OK);

	// Disconnected from host.
}

uint8_t usbd_testing(void)
{
	USBD_ctx usb_ctx;

	memset(&usb_ctx, 0, sizeof(usb_ctx));

	usb_ctx.standard_req_cb = NULL;
	usb_ctx.get_descriptor_cb = standard_req_get_descriptor;
	usb_ctx.class_req_cb = class_req_cb;
	usb_ctx.vendor_req_cb = vendor_req_cb;
	usb_ctx.suspend_cb = suspend_cb;
	usb_ctx.resume_cb = resume_cb;
	usb_ctx.reset_cb = reset_cb;
	usb_ctx.lpm_cb = NULL;
	usb_ctx.speed = USBD_SPEED_FULL;

	// Initialise the USB device with a control endpoint size
	// of 8 to 64 bytes. This must match the size for bMaxPacketSize0
	// defined in the device descriptor.
	usb_ctx.ep0_size = USB_CONTROL_EP_SIZE;
	//usb_ctx.ep0_cb = ep_cb;

	USBD_initialise(&usb_ctx);

	for (;;)
	{
		if (USBD_connect() == USBD_OK)
		{
			if (USBD_DFU_is_runtime())
			{
				// Start the CDC emulation code.
				cdc_uart_bridge();
			}
			else
			{
				// In DFU mode. Process USB requests.
				while (USBD_process() == USBD_OK);
			}
			tfp_printf("Restarting\r\n");
		}
	}

	return 0;
}

/* FUNCTIONS ***********************************************************************/

int main(void)
{
	/* Enable the UART Device... */
	sys_enable(sys_device_uart0);
	/* Make GPIO48 function as UART0_TXD and GPIO49 function as UART0_RXD... */
	gpio_function(48, pad_uart0_txd); /* UART0 TXD */
	gpio_function(49, pad_uart0_rxd); /* UART0 RXD */

	// Open the UART using the coding required.
	uartCoding();

	/* Print out a welcome message... */
	uart_puts(UART0,
			"\x1B[2J" /* ANSI/VT100 - Clear the Screen */
			"\x1B[H"  /* ANSI/VT100 - Move Cursor to Home */
	);

	/* Enable tfp_printf() functionality... */
	init_printf(UART0, tfp_putc);

	sys_enable(sys_device_timer_wdt);

	//interrupt_attach(interrupt_timers, (int8_t)interrupt_timers, ISR_timer);
	/* Enable the UART to fire interrupts when receiving data... */
	uart_enable_interrupt(UART0, uart_interrupt_rx);
	/* Enable the UART to fire interrupts when transmitting data... */
	uart_enable_interrupt(UART0, uart_interrupt_tx);
	/* Attach the interrupt so it can be called... */
	interrupt_attach(interrupt_uart0, (uint8_t) interrupt_uart0, uart0ISR);
	/* Enable interrupts to be fired... */
	uart_enable_interrupts_globally(UART0);

	/* Timer A = 1ms */
	timer_prescaler(1000);
	timer_init(timer_select_a, 100, timer_direction_down, timer_prescaler_select_on, timer_mode_continuous);
	timer_enable_interrupt(timer_select_a);
	timer_start(timer_select_a);

	interrupt_attach(interrupt_timers, (int8_t)interrupt_timers, ISR_timer);

	interrupt_enable_globally();

	tfp_printf("(C) Copyright 2016, Future Technology Devices International Ltd. \r\n");
	tfp_printf("--------------------------------------------------------------------- \r\n");
	tfp_printf("Welcome to USBDCDC ACM Tester Example 1... \r\n");
	tfp_printf("\r\n");
	tfp_printf("Emulate a CDC ACM device connected to the USB Device Port.\r\n");
	tfp_printf("--------------------------------------------------------------------- \r\n");

	usbd_testing();

	interrupt_detach(interrupt_timers);
	interrupt_disable_globally();

	sys_disable(sys_device_timer_wdt);

	// Wait forever...
	for (;;);

	return 0;
}


